from fastmcp import FastMCP, Context
import scvelo as scv
import scanpy as sc
import inspect
import os
from pathlib import Path
from ..schema.pp import *
from scmcp_shared.util import filter_args, add_op_log,forward_request,get_ads
from scmcp_shared.logging_config import setup_logger
from fastmcp.exceptions import ToolError
from scmcp_shared.schema import AdataModel

logger = setup_logger()

pp_mcp = FastMCP("CellrankMCP-Preprocessing-Server")


@pp_mcp.tool()
async def filter_and_normalize(
    request: FilterAndNormalizeModel,
    adinfo: AdataModel = AdataModel()
):
    """
    Preprocess data: filter and normalize AnnData object for velocity/pseudotime analysis.
    """
    try:
        result = await forward_request("filter_and_normalize", request, adinfo)
        if result is not None:
            return result   
        ads = get_ads()
        adata = ads.get_adata(adinfo=adinfo).copy()
        if 'spliced' in adata.layers:
            # Filter arguments based on the function parameters
            kwargs = filter_args(request, scv.pp.filter_and_normalize)
            if "log" in inspect.signature(scv.pp.filter_and_normalize).parameters:
                kwargs["log"] = False
            # Run filter_and_normalize
            scv.pp.filter_and_normalize(adata, **kwargs)            
            sc.pp.log1p(adata)
            add_op_log(adata, scv.pp.filter_and_normalize, kwargs, adinfo)
            add_op_log(adata, sc.pp.log1p, {}, adinfo)
        else:
            kwargs = filter_args(request, sc.pp.filter_genes)
            sc.pp.filter_genes(adata, filter_args(request, sc.pp.filter_genes))
            sc.pp.normalize_total(adata, filter_args(request, sc.pp.normalize_total))
            sc.pp.log1p(adata)
            sc.pp.highly_variable_genes(adata,  filter_args(request, sc.pp.highly_variable_genes))
            adata.layers["spliced"] = adata.X
            adata.layers["unspliced"] = adata.X
        ads.set_adata(adata, adinfo=adinfo)
        return {"sampleid": adinfo.sampleid or ads.active_id, "dtype": adinfo.adtype, "adata": adata}
        
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, '__context__') and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)




import asyncio
import os
import sys
import typer
from enum import Enum
from typing import Optional
from scmcp_shared.util import add_figure_route

app = typer.Typer(
    name="cellrankmcp",
    help="Cellrank MCP Server CLI",
    add_completion=False,
    no_args_is_help=True,  # Show help if no args provided    
    )

# Define enums for choices
class Module(str, Enum):
    ALL = "all"


class Transport(str, Enum):
    STDIO = "stdio"
    SSE = "sse"
    SHTTP = "shttp"



@app.command(name="run")
def run(
    log_file: Optional[str] = typer.Option(None, "--log-file", help="Log file path, use stdout if None"),
    module: Module = typer.Option(Module.ALL, "-m", "--module", help="Specify modules to load", 
                              case_sensitive=False),
    transport: Transport = typer.Option(Transport.STDIO, "-t", "--transport", help="Specify transport type", 
                                 case_sensitive=False),
    port: int = typer.Option(8000, "-p", "--port", help="transport port"),
    host: str = typer.Option("127.0.0.1", "--host", help="transport host"),
    forward: str = typer.Option(None, "-f", "--forward", help="forward request to another server"),
):
    """Start cellrank MCP Server"""
    
    if log_file is not None:
        os.environ['SCMCP_LOG_FILE'] = log_file
    if forward is not None:
        os.environ['SCMCP_FORWARD'] = forward
    
    os.environ['SCMCP_TRANSPORT'] = transport.value
    os.environ['SCMCP_HOST'] = host
    os.environ['SCMCP_PORT'] = str(port)
    os.environ['SCMCP_MODULE'] = module.value
    
    from .server import cellrank_mcp, setup
    asyncio.run(setup())

    if transport == Transport.STDIO:
        cellrank_mcp.run()
    elif transport == Transport.SSE:
        add_figure_route(cellrank_mcp)
        cellrank_mcp.run(
                transport="sse",
                host=host, 
                port=port, 
                log_level="info"
            )

    elif transport == Transport.SHTTP:
        add_figure_route(cellrank_mcp)
        cellrank_mcp.run(
                transport="streamable-http",
                host=host, 
                port=port, 
                log_level="info"
            )


@app.callback()
def main():
    """Cellrank MCP CLI root command."""
    pass


from fastmcp import FastMCP, Context
from fastmcp.exceptions import ToolError
import os
import inspect
from pathlib import Path
import scanpy as sc
import cellrank as cr
import scvelo as scv
from pathlib import Path
from ..schema.kernel import *
from scmcp_shared.util import filter_args,forward_request, get_ads
from scmcp_shared.logging_config import setup_logger
from scmcp_shared.schema import AdataModel

logger = setup_logger()


kernel_mcp = FastMCP("CellrankMCP-Kernel-Server")


@kernel_mcp.tool()
async def create_kernel(
    request: KernelModel, 
    adinfo: AdataModel = AdataModel()
):
    """Create a CellRank kernel based on the specified type and parameters."""
    try:
        # Check if AnnData object exists in the session
        result = await forward_request("kernel_create_kernel", request, adinfo)
        if result is not None:
            return result
        ads = get_ads()
        adata = ads.get_adata(adinfo=adinfo).copy()
        kernel_type = request.kernel
        
        kernel_dic = {
            "pseudotime": cr.kernels.PseudotimeKernel,
            "cytotrace": cr.kernels.CytoTRACEKernel,
            "velocity": cr.kernels.VelocityKernel,
            "connectivity": cr.kernels.ConnectivityKernel,
            "realtime": cr.kernels.RealTimeKernel
        }
        if kernel_type not in kernel_dic:
            return {
                "status": "error",
                "message": f"Unsupported kernel type: {kernel_type}"
            }
        kernel_class = kernel_dic[kernel_type]
        kwargs = filter_args(request, kernel_class)
        if kernel_type == "cytotrace":
            if "Ms" not in adata.layers or "Mu" not in adata.layers:
                scv.pp.moments(adata, n_pcs=30, n_neighbors=30)
            kernel = kernel_class(adata=adata, **kwargs).compute_cytotrace()
        else:
            kernel = kernel_class(adata=adata, **kwargs)
        
        ads.cr_kernel[kernel_type] = kernel
        ads.set_adata(adata, adinfo=adinfo)
        return {
            "status": "success",
            "message": f"Successfully created {kernel_type} kernel",
            "kernel_type": kernel_type
        }
    
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, '__context__') and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)

@kernel_mcp.tool()
async def compute_transition_matrix(
    request: ComputeTransitionMatrixModel,
    adinfo: AdataModel = AdataModel()
    ):
    """Compute transition matrix for a specified kernel using appropriate parameters."""
    try:
        result = await forward_request("compute_transition_matrix", request, adinfo)
        if result is not None:
            return result
        kernel_type = request.kernel
        ads = get_ads()
        kernel = ads.cr_kernel[kernel_type].copy()        
        kwargs = filter_args(request, kernel.compute_transition_matrix)
        kwargs["show_progress_bar"] = False
        kernel.compute_transition_matrix(**kwargs)
        ads.cr_kernel[kernel_type] = kernel
        return {
            "status": "success",
            "message": f"Successfully computed transition matrix for {kernel_type} kernel"
        }
        
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, '__context__') and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)


from fastmcp import FastMCP, Context
from fastmcp.exceptions import ToolError
import inspect
from pathlib import Path
import os
from ..schema.pl import *
import scvelo as scv
import cellrank as cr
from scmcp_shared.util import add_op_log, filter_args,forward_request,get_ads
from scmcp_shared.logging_config import setup_logger
from ..util import set_fig_path
from scmcp_shared.schema import AdataModel
logger = setup_logger()


pl_mcp = FastMCP("CellrankMCP-Kernel-Server")


@pl_mcp.tool()
async def kernel_projection(
    request: KernelPlotProjectionModel, 
    adinfo: AdataModel = AdataModel()
    ):
    """Plot transition matrix as a stream or grid plot for a specified kernel."""
    try:
        result = await forward_request("kernel_projection", request, adinfo)
        if result is not None:
            return result 
        kernel_type = request.kernel
        ads = get_ads()
        kernel = ads.cr_kernel[kernel_type]
        
        # Check if transition matrix has been computed
        if not hasattr(kernel, 'transition_matrix') or kernel.transition_matrix is None:
            error_msg = f"Transition matrix for kernel '{kernel_type}' has not been computed. Please compute transition matrix first."
            logger.error(error_msg)
            raise ValueError(error_msg)
        
        # Filter arguments based on the plot_projection method
        kwargs = filter_args(request, kernel.plot_projection)
        kwargs["save"] = kernel_type
        kwargs["show"] = False
        kwargs["color"] = request.color
        kwargs["legend_loc"] = request.legend_loc
        # Plot the projection
        kernel.plot_projection(**kwargs)
        kwargs["kernel"] = kernel_type
        fig_path = set_fig_path("scvelo_projection", **kwargs)
        return {"figpath": fig_path}
        
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, '__context__') and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)


@pl_mcp.tool()
async def circular_projection(
    request: CircularProjectionModel,
    adinfo: AdataModel = AdataModel()
):
    """
    Visualize fate probabilities in a circular embedding. compute_fate_probabilities first.
    """
    try:
        result = await forward_request("circular_projection", request, adinfo)
        if result is not None:
            return result   
        # Check if AnnData object exists in the session
        ads = get_ads()
        adata = ads.get_adata(adinfo=adinfo)

        if "lineages_fwd" not in adata.obsm:
            raise ValueError("No lineages_fwd found. Please call compute_fate_probabilities first.")
        
        # Check if estimator exists and has computed fate probabilities
        kernel_types = []
        kernel_types = list(ads.cr_estimator.keys())
                
        # Filter arguments based on the circular_projection function
        kwargs = filter_args(request, cr.pl.circular_projection)
        
        kwargs["save"] = "circular_projection.png"
        kwargs["legend_loc"] = "right"
        kwargs["keys"] = request.keys
        # Call the circular_projection function
        cr.pl.circular_projection(adata, **kwargs)
        del kwargs["save"]
        fig_path = set_fig_path("circular_projection", **kwargs)
        return {
            "status": "success",
            "message": "Successfully created circular projection plot",
            "figpath": fig_path
        }
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, '__context__') and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)




import asyncio
from fastmcp import FastMCP
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Any


from .pl import pl_mcp
from .kernel import kernel_mcp
from .pp import pp_mcp
from .estimator import estimator_mcp

import scmcp_shared.server as shs
from  scmcp_shared.util import filter_tools

ads = shs.AdataState()

@asynccontextmanager
async def adata_lifespan(server: FastMCP) -> AsyncIterator[Any]:
    yield ads

cellrank_mcp = FastMCP("Cellrank-MCP-Server", lifespan=adata_lifespan)

async def setup(modules=None):

    pp1_mcp = await filter_tools(
        shs.pp_mcp,
        include_tools=["neighbors"]
    )
    mcp_dic = {
        "io": shs.io_mcp, 
        "pp": pp1_mcp, 
        "kernel": kernel_mcp, 
        "estimator": estimator_mcp, 
        "pl": shs.pl_mcp, "tl": shs.tl_mcp
        }
    if modules is None or modules == "all":
        modules = mcp_dic.keys()
    for module in modules:
        await cellrank_mcp.import_server(module, mcp_dic[module])
    await cellrank_mcp.import_server("pl", pl_mcp)
    await cellrank_mcp.import_server("pp", pp_mcp)

